import { AxiosError } from "axios";
import { ApiError } from "./ApiError";
import { useAxios } from "./useAxios";
import { MutateConfigOptions } from "./MutateConfigOptions";
import { BaseQueryKey, QueryBuilder } from "./QueryBuilder";
import { useMutation, useQueryClient } from "@tanstack/vue-query";

/**
 * A wrapper around useMutation that handles the common use case of creating a new resource.
 */
export class Mutate<QueryKey extends BaseQueryKey, T> {
  config: MutateConfigOptions;

  defaultConfig: MutateConfigOptions = {
    autoInvalidate: true,
  };

  withDefaults(config?: Partial<MutateConfigOptions>) {
    return { ...this.defaultConfig, ...(config ?? {}) };
  }

  constructor(config?: Partial<MutateConfigOptions>) {
    this.config = this.withDefaults(config);
  }

  create<Vars, Return = unknown>({
    builderFunc,
    invalidate,
    select,
  }: {
    select?: (data: T) => Return;
    /**
     * Function that builds the query builder.
     *
     * @param data The data passed to the mutation.
     * @param data.vars The variables passed to the mutation.
     * @param data.config The configuration passed to the mutation.
     * @returns Instance of QueryBuilder
     */
    builderFunc: (data: {
      vars: Vars;
      config: MutateConfigOptions;
    }) => QueryBuilder<QueryKey>;
    invalidate?: (data: {
      vars: Vars;
      config: MutateConfigOptions;
    }) => QueryKey[][];
  }) {
    const { axios } = useAxios();

    const queryClient = useQueryClient();

    const triggerInvalidate = async (vars: Vars) => {
      if (invalidate === undefined) return;

      const keys = invalidate({ vars, config: this.config });

      keys?.forEach(async (key) => {
        await queryClient.invalidateQueries({
          queryKey: key,
        });
      });
    };

    const mutate = useMutation<Return, AxiosError<ApiError>, Vars>({
      mutationFn: async (vars) => {
        const queryBuilder = builderFunc({ vars, config: this.config });

        const data = await axios
          .post(queryBuilder.path.value, queryBuilder.body.value, {
            params: queryBuilder.params.value,
          })
          .then((res) => (select ? select(res.data) : res.data));

        return data;
      },

      onSuccess: async (_data, vars) => {
        if (this.config.autoInvalidate) triggerInvalidate(vars);
      },
    });

    return { ...mutate, triggerInvalidate };
  }

  update<Vars, Return = unknown>({
    builderFunc,
    invalidate,
  }: {
    builderFunc: (data: {
      vars: Vars;
      config: MutateConfigOptions;
    }) => QueryBuilder<QueryKey>;
    invalidate?: (data: {
      vars: Vars;
      config: MutateConfigOptions;
    }) => QueryKey[][];
  }) {
    const { axios } = useAxios();

    const queryClient = useQueryClient();

    const triggerInvalidate = async (vars: Vars) => {
      if (invalidate === undefined) return;

      const keys = invalidate({ vars, config: this.config });

      keys?.forEach(async (key) => {
        await queryClient.invalidateQueries({
          queryKey: key,
        });
      });
    };

    return useMutation<Return, AxiosError, Vars>({
      mutationFn: async (vars) => {
        const queryBuilder = builderFunc({ vars, config: this.config });

        const data = await axios
          .put(queryBuilder.path.value, queryBuilder.body.value, {
            params: queryBuilder.params.value,
          })
          .then((res) => res.data);
        return data;
      },

      onSuccess: async (_data, vars) => {
        if (this.config.autoInvalidate) triggerInvalidate(vars);
      },
    });
  }

  delete<Vars, Return = unknown>({
    builderFunc,
    invalidate,
  }: {
    builderFunc: (data: {
      vars: Vars;
      config: MutateConfigOptions;
    }) => QueryBuilder<QueryKey>;
    invalidate?: (data: {
      vars: Vars;
      config: MutateConfigOptions;
    }) => QueryKey[][];
  }) {
    const { axios } = useAxios();

    const queryClient = useQueryClient();

    const triggerInvalidate = async (vars: Vars) => {
      if (invalidate === undefined) return;

      const keys = invalidate({ vars, config: this.config });

      keys?.forEach(async (key) => {
        await queryClient.invalidateQueries({
          queryKey: key,
        });
      });
    };

    return useMutation<Return, AxiosError, Vars>({
      mutationFn: async (vars) => {
        const queryBuilder = builderFunc({ vars, config: this.config });

        const data = await axios
          .delete(queryBuilder.path.value, queryBuilder.body.value)
          .then((res) => res.data);
        return data;
      },

      onSuccess: async (_data, vars) => {
        if (this.config.autoInvalidate) triggerInvalidate(vars);
      },
    });
  }
}
